@@ -35,8 +35,8 @@ extern void die(const char *,struct pt_regs *,long);
 
 static ATOMIC_NOTIFIER_HEAD(notify_page_fault_chain);
 
-int suspend2_faulted = 0;
-EXPORT_SYMBOL(suspend2_faulted);
+int toi_faulted = 0;
+EXPORT_SYMBOL(toi_faulted);
 
 int register_page_fault_notifier(struct notifier_block *nb)
 {
@@ -288,6 +288,8 @@ static inline int vmalloc_fault(unsigned long address)
 	return 0;
 }
 
+int show_unhandled_signals = 1;
+
 /*
  * This routine handles page faults.  It determines the address,
  * and the problem, and then passes it off to one of the appropriate
@@ -308,6 +310,7 @@ fastcall void __kprobes do_page_fault(struct pt_regs *regs,
 	struct vm_area_struct * vma;
 	unsigned long address;
 	int write, si_code;
+	int fault;
 
 	/* get the address */
         address = read_cr2();
@@ -316,19 +319,21 @@ fastcall void __kprobes do_page_fault(struct pt_regs *regs,
 
 	si_code = SEGV_MAPERR;
 
-	/* During a Suspend2 atomic copy, with DEBUG_SLAB, we will
+	/* During a TuxOnIce atomic copy, with DEBUG_SLAB, we will
 	 * get page faults where slab has been unmapped. Map them
-	 * temporarily and set the variable that tells Suspend2 to
+	 * temporarily and set the variable that tells TuxOnIce to
 	 * unmap afterwards.
 	 */
 
-	if (unlikely(suspend2_running && !suspend2_faulted)) {
+#ifdef CONFIG_DEBUG_PAGEALLOC
+	if (unlikely(toi_running && !toi_faulted)) {
 		struct page *page = NULL;
-		suspend2_faulted = 1;
+		toi_faulted = 1;
 		page = virt_to_page(address);
 		kernel_map_pages(page, 1, 1);
 		return;
 	}
+#endif
 
 	/*
 	 * We fault-in kernel-space virtual memory on-demand. The
@@ -441,20 +446,18 @@ good_area:
 	 * make sure we exit gracefully rather than endlessly redo
 	 * the fault.
 	 */
-	switch (handle_mm_fault(mm, vma, address, write)) {
-		case VM_FAULT_MINOR:
-			tsk->min_flt++;
-			break;
-		case VM_FAULT_MAJOR:
-			tsk->maj_flt++;
-			break;
-		case VM_FAULT_SIGBUS:
-			goto do_sigbus;
-		case VM_FAULT_OOM:
+	fault = handle_mm_fault(mm, vma, address, write);
+	if (unlikely(fault & VM_FAULT_ERROR)) {
+		if (fault & VM_FAULT_OOM)
 			goto out_of_memory;
-		default:
-			BUG();
+		else if (fault & VM_FAULT_SIGBUS)
+			goto do_sigbus;
+		BUG();
 	}
+	if (fault & VM_FAULT_MAJOR)
+		tsk->maj_flt++;
+	else
+		tsk->min_flt++;
 
 	/*
 	 * Did it hit the DOS screen memory VA from vm86 mode?
@@ -489,6 +492,14 @@ bad_area_nosemaphore:
 		if (is_prefetch(regs, address, error_code))
 			return;
 
+		if (show_unhandled_signals && unhandled_signal(tsk, SIGSEGV) &&
+		    printk_ratelimit()) {
+			printk("%s%s[%d]: segfault at %08lx eip %08lx "
+			    "esp %08lx error %lx\n",
+			    tsk->pid > 1 ? KERN_INFO : KERN_EMERG,
+			    tsk->comm, tsk->pid, address, regs->eip,
+			    regs->esp, error_code);
+		}
 		tsk->thread.cr2 = address;
 		/* Kernel addresses are always protection faults */
 		tsk->thread.error_code = error_code | (address >= TASK_SIZE);
