@@ -8,8 +8,8 @@
 
 static inline int arch_prepare_suspend(void) { return 0; }
 
-extern int suspend2_faulted;
-#define clear_suspend2_fault() do { suspend2_faulted = 0; } while(0)
+extern int toi_faulted;
+#define clear_toi_fault() do { toi_faulted = 0; } while(0)
 
 /* image of the saved processor state */
 struct saved_context {
@@ -24,7 +24,7 @@ struct saved_context {
 	unsigned long return_address;
 } __attribute__((packed));
 
-#ifdef CONFIG_ACPI_SLEEP
+#ifdef CONFIG_ACPI
 extern unsigned long saved_eip;
 extern unsigned long saved_esp;
 extern unsigned long saved_ebp;
