@@ -33,7 +33,7 @@
 #include <asm/io.h>
 
 #include "power.h"
-#include "suspend2_builtin.h"
+#include "tuxonice_builtin.h"
 
 static int swsusp_page_is_free(struct page *);
 static void swsusp_set_page_forbidden(struct page *);
@@ -48,9 +48,9 @@ struct pbe *restore_pblist;
 int resume_attempted;
 EXPORT_SYMBOL_GPL(resume_attempted);
 
-#ifdef CONFIG_SUSPEND2
-#include "pagedir.h"
-int suspend_post_context_save(void);
+#ifdef CONFIG_TOI
+#include "tuxonice_pagedir.h"
+int toi_post_context_save(void);
 #endif
 
 /* Pointer to an auxiliary buffer (1 page) */
@@ -94,9 +94,9 @@ static void *get_image_page(gfp_t gfp_mask, int safe_needed)
 
 unsigned long get_safe_page(gfp_t gfp_mask)
 {
-#ifdef CONFIG_SUSPEND2
-	if (suspend2_running)
-		return suspend_get_nonconflicting_page();
+#ifdef CONFIG_TOI
+	if (toi_running)
+		return toi_get_nonconflicting_page();
 #endif
 
 	return (unsigned long)get_image_page(gfp_mask, PG_SAFE);
@@ -1215,9 +1215,9 @@ asmlinkage int swsusp_save(void)
 {
 	unsigned int nr_pages, nr_highmem;
 
-#ifdef CONFIG_SUSPEND2
-	if (suspend2_running)
-		return suspend_post_context_save();
+#ifdef CONFIG_TOI
+	if (toi_running)
+		return toi_post_context_save();
 #endif
 
 	printk("swsusp: critical section: \n");
